/*
 * Does the required checks for the files and also provides function
 * for printing the violations.
 *
 * Copyright (c) 2004 Aki Tossavainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include "eics.h"

/* Check for violations */
int
checkViolations (const struct hash_entry_s *original,
		 const struct hash_entry_s *current, int policy)
{
  int violations = 0;
  if ((policy & POLICY_ACCESS_TIME) &&
      (original->filestat.st_atime != current->filestat.st_atime))
    {
      violations |= POLICY_ACCESS_TIME;
    }
  if ((policy & POLICY_CREATION_TIME) &&
      (original->filestat.st_ctime != current->filestat.st_ctime))
    {
      violations |= POLICY_CREATION_TIME;
    }
  if ((policy & POLICY_MODIFY_TIME) &&
      (original->filestat.st_mtime != current->filestat.st_mtime))
    {
      violations |= POLICY_MODIFY_TIME;
    }
  if ((policy & POLICY_GROUP) &&
      (original->filestat.st_gid != current->filestat.st_gid))
    {
      violations |= POLICY_GROUP;
    }
  if ((policy & POLICY_INODE) &&
      (original->filestat.st_ino != current->filestat.st_ino))
    {
      violations |= POLICY_INODE;
    }
  if ((policy & POLICY_SIZE) &&
      (original->filestat.st_size != current->filestat.st_size))
    {
      if ((policy & POLICY_MAYGROW) &&
	  (original->filestat.st_size > current->filestat.st_size))
	{
	  violations |= POLICY_SIZE;
	}
      else if (!(policy & POLICY_MAYGROW))
	{
	  violations |= POLICY_SIZE;
	}
    }
  if ((policy & POLICY_PERMISSIONS) &&
      (original->filestat.st_mode != current->filestat.st_mode))
    {
      violations |= POLICY_PERMISSIONS;
    }
  if ((policy & POLICY_OWNER) &&
      (original->filestat.st_uid != current->filestat.st_uid))
    {
      violations |= POLICY_OWNER;
    }
  if ((policy & POLICY_NLINKS) &&
      (original->filestat.st_nlink != current->filestat.st_nlink))
    {
      violations |= POLICY_NLINKS;
    }
  if ((policy & POLICY_MESH) &&
      ((memcmp
	(original->meshes.sha1, current->meshes.sha1,
	 sizeof (current->meshes.sha1)))
       ||
       (memcmp
	(original->meshes.ripemd, current->meshes.ripemd,
	 sizeof (current->meshes.ripemd)))))
    {
      violations |= POLICY_MESH;
    }
  return violations;
}

/* Prints all found violations */
void
printViolations (const struct hash_entry_s *original,
		 const struct hash_entry_s *current, int violations)
{
  if (violations & POLICY_ACCESS_TIME)
    {
      fprintf (stderr, "        Access time has changed(%lu->%lu)\n",
	       original->filestat.st_atime, current->filestat.st_atime);
    }
  if (violations & POLICY_CREATION_TIME)
    {
      fprintf (stderr, "       Creation time has changed(%lu->%lu)\n",
	       original->filestat.st_ctime, current->filestat.st_ctime);
    }
  if (violations & POLICY_MODIFY_TIME)
    {
      fprintf (stderr, "       Modification time has changed(%lu->%lu)\n",
	       original->filestat.st_mtime, current->filestat.st_mtime);
    }
  if (violations & POLICY_GROUP)
    {
      fprintf (stderr, "       Owning group ID has changed(%u->%u)\n",
	       original->filestat.st_gid, current->filestat.st_gid);
    }
  if (violations & POLICY_OWNER)
    {
      fprintf (stderr, "       Owning user ID has changed(%u->%u)\n",
	       original->filestat.st_uid, current->filestat.st_uid);
    }
  if (violations & POLICY_INODE)
    {
#if defined(HAVE_INO64_T)
      fprintf (stderr, "       Entry inode has changed(%llu->%llu)\n",
	       original->filestat.st_ino, current->filestat.st_ino);
#else
      fprintf (stderr, "       Entry inode has changed(%lu->%lu)\n",
	       original->filestat.st_ino, current->filestat.st_ino);
#endif
    }
  if (violations & POLICY_PERMISSIONS)
    {
      if ((original->filestat.st_mode & ~S_IFMT) !=
	  (current->filestat.st_mode & ~S_IFMT))
	{
	  fprintf (stderr,
		   "       Entry permissions have changed(%04o->%04o)\n",
		   original->filestat.st_mode & ~S_IFMT,
		   current->filestat.st_mode & ~S_IFMT);
	}
      if ((original->filestat.st_mode & S_IFMT) !=
	  (current->filestat.st_mode & S_IFMT))
	{
	  fprintf (stderr, "       Entry type has changed(%u->%u)\n",
		   original->filestat.st_mode & S_IFMT,
		   current->filestat.st_mode & S_IFMT);
	}
    }
  if (violations & POLICY_NLINKS)
    {
      fprintf (stderr,
	       "       Number of links from entry has changed(%u->%u)\n",
	       original->filestat.st_nlink, current->filestat.st_nlink);
    }
  if (violations & POLICY_SIZE)
    {
#if defined(HAVE_OFF64_T)
      fprintf (stderr, "       Entry size has changed(%llu->%llu)\n",
	       original->filestat.st_size, current->filestat.st_size);
#else
      fprintf (stderr, "       Entry size has changed(%lu->%lu)\n",
	       original->filestat.st_size, current->filestat.st_size);
#endif
    }
  if (violations & POLICY_MESH)
    {
      fprintf (stderr, "       Entry contents have changed\n");
    }
}
