/*
 * Main inclusion file, used for all files
 *
 * (c) Aki Tossavainen 2004

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */
#ifndef EICS_H
#define EICS_H 1

/* If you have a VERY VERY BIG filesystem, you can increase this value
 * to make the scan faster. Or if your filesystem is considerably smaller
 * you can decrease this to use less memory. At the moment it's 2^17
 * and will eat at least 500k of memory. (or 1 meg on 64 bit systems)
 */
#define HASHTABLE_SIZE 131072

#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/evp.h>
#include <openssl/sha.h>
#include <openssl/ripemd.h>
#include <openssl/err.h>

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif
#include <zlib.h>
#include <pcre.h>
#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif
#include <unistd.h>
#if STDC_HEADERS
# include <string.h>
#else
# if !HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# if !HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#ifndef S_ISLNK
#define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
#endif

#define EICS_DB_FILE DATABASEDIR "/eics.db"
#define EICS_DB_FILE_TMP EICS_DB_FILE ".tmp"
#define EICS_DB_FILE_SIG EICS_DB_FILE ".sig"
#define EICS_CONFIG_FILE CONFIGDIR "/eics.conf"
#define EICS_CONFIG_FILE_SIG EICS_CONFIG_FILE ".sig"
#define EICS_PUB_KEY CONFIGDIR "/eics.key.pub"
#define EICS_PRIV_KEY CONFIGDIR "/eics.key.priv"

#define POLICY_ACCESS_TIME 0x1
#define POLICY_CREATION_TIME 0x2
#define POLICY_MODIFY_TIME 0x4
#define POLICY_GROUP 0x8
#define POLICY_MESH 0x10
#define POLICY_INODE 0x20
#define POLICY_MAYGROW 0x40
#define POLICY_PERMISSIONS 0x80
#define POLICY_NLINKS 0x100
#define POLICY_OWNER 0x200
#define POLICY_SIZE 0x400
#define POLICY_EXCLUDE 0x800
#define POLICY_NOSUBDIRS 0x1000

#define DSA_KEY_STRENGTH 1024

#define MINPASSLEN 8

#define DBMAGICNUM 31545

/* Meshes for checking the file contents */
struct mesh_s
{
  unsigned char sha1[RIPEMD160_DIGEST_LENGTH];
  unsigned char ripemd[SHA_DIGEST_LENGTH];
};

/* Filesystem entry for checking */
struct hash_entry_s
{
  int checked;
  struct stat filestat;
  struct mesh_s meshes;
  unsigned int namelen;
  unsigned int hashval;
  char *name;
  /* pointer to collision chain next */
  struct hash_entry_s *next;
};

void initHash (void);
void updateEntryToHashTable (const struct hash_entry_s *newEntry);
struct hash_entry_s *findEntryFromHashTable (const char *entry);
int writeDatabase (const char *dbfile, const char *tmpfile,
		   const char *sigfile);
int loadDatabase (const char *dbfile, const char *sigfile);
int pruneDatabase (void);
int entryChecked (const char *name);

int processDirectory (const char *directory, int defaultPolicy);

int meshFile (const char *file, struct mesh_s *mesh);

void initPolicy (void);
int matchPolicy (const char *file, size_t flen, int oldpolicy);
int loadPolicy (const char *policyfile, const char *policysigfile,
		int doUpdate);
int getFirstPolicyFlags (void);
const char *getFirstPolicyDir (void);

int writeDSAKeyPair (DSA * dsa, const char *publickeyfile,
		     const char *privatekeyfile);
int generateNewKeys (const char *publickeyfile, const char *privatekeyfile);
int checkSignature (const char *file, const char *sigfile,
		    const char *pubkeyfile);
int writeSignature (const char *file, const char *sigfile,
		    const char *privkeyfile);

int pass_cb (char *pass, int maxlen, int rwflag, void *u);

int checkViolations (const struct hash_entry_s *original,
		     const struct hash_entry_s *current, int policy);
void printViolations (const struct hash_entry_s *original,
		      const struct hash_entry_s *current, int violations);

char *mb_strrchr (char *haystack, char needle);
char *mb_strchr (char *haystack, char needle);

#endif
