/*
 * File hash table, we use this to speed up things alot. Now we don't have 
 * to roam thru a linked list every time, but instead we can get access times
 * that are near O(1). Compared to old one which was guaranteed to be O(n)...
 * 
 * hash values are stored during database save to speed up load process.
 *
 * Copyright (c) 2004 Aki Tossavainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include "eics.h"

/* Our hash table */
struct hash_table_s
{
  struct hash_entry_s *entries[HASHTABLE_SIZE];
#ifdef HAVE_UNSIGNED_LONG_LONG
  unsigned long long nentry;
#else
  unsigned long nentry;
#endif
} hash_table;

void
initHash (void)
{
  memset (hash_table.entries, 0, sizeof (hash_table.entries));
  hash_table.nentry = 0;
}

/* Checks if entry is checked */
int
entryChecked (const char *name)
{
  struct hash_entry_s *entry;
  if ((entry = findEntryFromHashTable (name)))
    {
      return entry->checked;
    }
  return 0;
}

/*
 * We loan the crc32 from libz
 */
unsigned long
calculateHash (const char *name, unsigned int length)
{
  unsigned long crc;
  /* get initial value */
  crc = crc32 (0L, Z_NULL, 0);
  /* calculate CRC */
  crc = crc32 (crc, (const unsigned char *) name, length);
  /* return hash value */
  return crc % HASHTABLE_SIZE;
}

/* Looks up a hash entry from the hash chains */
struct hash_entry_s *
findHashEntry (unsigned int hashval, const char *name,
	       struct hash_entry_s **stop)
{
  /* take the correct collision chain */
  struct hash_entry_s *hash_entry = hash_table.entries[hashval];
  /* as long as we have entries */
  if (stop)
    *stop = NULL;
  while (hash_entry)
    {
      if (stop)
	*stop = hash_entry;
      if (!strcmp (hash_entry->name, name))
	return hash_entry;
      /* update stop pointer, this reduces time if entry not found 
       * when we perform addition in the next phase.
       */
      hash_entry = hash_entry->next;
    }
  /* nothing found */
  return NULL;
}

struct hash_entry_s *
findEntryFromHashTable (const char *entry)
{
  return findHashEntry (calculateHash (entry, strlen (entry)), entry, NULL);
}

/* Copies entry to another, note that it is not possible to copy 
 * the entry directly
 */
void
copyEntry (struct hash_entry_s *dest, const struct hash_entry_s *source)
{
  dest->checked = 1;
  /* Statistics */
  memcpy (&dest->filestat, &source->filestat, sizeof (source->filestat));
  /* sha1 + ripemd */
  memcpy (&dest->meshes, &source->meshes, sizeof (source->meshes));
  /* Entry name */
  dest->name = malloc (sizeof (char) * (source->namelen + 1));
  memset (dest->name, 0, sizeof (char) * (source->namelen + 1));
  strncpy (dest->name, source->name, source->namelen);
  /* Name length */
  dest->namelen = source->namelen;
  /* Hash value */
  dest->hashval = source->hashval;
}

/* updates entry in hash table */
void
updateEntryToHashTable (const struct hash_entry_s *newEntry)
{
  /* some pointers we need */
  struct hash_entry_s *entry, *stop;
  /* hash value */
  unsigned int hashval;
  /* get hash val */
  hashval = calculateHash (newEntry->name, newEntry->namelen);
  /* init stop */
  stop = NULL;
  /* look if entry is already there */
  if ((entry = findHashEntry (hashval, newEntry->name, &stop)))
    {
      /* update 
       * NB! the 'next' pointer is not modified.
       */
      copyEntry (entry, newEntry);
      entry->hashval = hashval;
    }
  else
    {
      /* or create new */
      entry = malloc (sizeof (struct hash_entry_s));
      memset (entry, 0, sizeof (struct hash_entry_s));
      copyEntry (entry, newEntry);
      /* hash val is not known in advance... */
      entry->hashval = hashval;
      /* update hash table chain */
      if (!hash_table.entries[hashval])
	{
	  hash_table.entries[hashval] = entry;
	}
      /* now you see why stop was collected in findHashEntry 
       * We do not need to look it up again, as we already have it
       */
      if (stop)
	{
	  stop->next = entry;
	}
      hash_table.nentry++;
    }
}

/* Writes the database into file 
 * The database is compressed using gzip to reduce it's size
 */
int
writeDatabase (const char *dbfile, const char *tmpfile, const char *sigfile)
{
  struct stat statent;
  int i;
  struct header_s
  {
    unsigned int magic;
    unsigned long timestamp;
#ifdef HAVE_UNSIGNED_LONG_LONG
    unsigned long long nentry;
#else
    unsigned long nentry;
#endif
  } header;
  struct write_entry_s
  {
    int checked;
    struct stat filestat;
    struct mesh_s meshes;
    unsigned int namelen;
    unsigned int hashval;
  } writebuf;
  struct hash_entry_s *ptr;
  gzFile fout;

  /* Create temp file, we store the
   * database into temporary file first to avoid signature problems
   * and avoid loading the private key before it's actually necessary
   */
  if (!(fout = gzopen (tmpfile, "wb")))
    {
      fprintf (stderr, "Unable to open %s for writing\n", tmpfile);
      return -1;
    }

  /* Make sure that only we can read it (or root...) */
  chmod (tmpfile, S_IRUSR | S_IWUSR);

  /* Setup header */
  header.magic = DBMAGICNUM;
  header.timestamp = time ((time_t *) NULL);
  header.nentry = hash_table.nentry;

  /* Write header */
  if (gzwrite (fout, &header, sizeof (header)) != sizeof (header))
    {
      fprintf (stderr, "Unable to write into %s\n", tmpfile);
      gzclose (fout);
      unlink (tmpfile);
      return -1;
    }

  /* Go thru all hash buckets and write all chains */
  for (i = 0; i < HASHTABLE_SIZE; i++)
    {
      if ((ptr = hash_table.entries[i]))
	{
	  while (ptr)
	    {
	      /* Copy only partial of the actual header to write
	       * since name is only a pointer 
	       */
	      memcpy (&writebuf, ptr, sizeof (writebuf));
	      /* Write it */
	      if (gzwrite (fout, &writebuf, sizeof (writebuf)) !=
		  sizeof (writebuf))
		{
		  fprintf (stderr, "Unable to write into %s\n", tmpfile);
		  gzclose (fout);
		  unlink (tmpfile);
		  return -1;
		}
	      /* Now write the name right after it, only the lenght required
	       * This a: saves space b: is more sane
	       */
	      if (gzwrite
		  (fout, ptr->name,
		   sizeof (char) * (ptr->namelen + 1)) !=
		  sizeof (char) * (ptr->namelen + 1))
		{
		  fprintf (stderr, "Unable to write into %s\n", tmpfile);
		  gzclose (fout);
		  unlink (tmpfile);
		  return -1;
		}
	      if (header.nentry)
		{
		  header.nentry--;
		}
	      else
		{
		  fprintf (stderr, "Ran out of entries\n");
		  gzclose (fout);
		  unlink (tmpfile);
		  return -1;
		}
	      ptr = ptr->next;
	    }
	}
    }
  gzclose (fout);
  if (header.nentry) 
    {
      fprintf (stderr, "All entries not written\n");
      unlink (tmpfile);
      return -1;
    }
  if (writeSignature (tmpfile, sigfile, EICS_PRIV_KEY))
    {
      unlink (tmpfile);
      return -1;
    }
  /* Is dbfile existent? */
  if (!lstat (dbfile, &statent))
    {
      /* If so, unlink it. This removes any symlink shit too */
      if (unlink (dbfile))
	{
	  /* Couldn't remove it, won't do anything */
	  fprintf (stderr, "Unable to remove old database, bailing out\n");
	  unlink (sigfile);
	  unlink (tmpfile);
	  return -1;
	}
    }
  rename (tmpfile, dbfile);
  /* Ensure database file is ours only */
  chmod (tmpfile, S_IRUSR | S_IWUSR);
  return 0;
}

int
loadDatabase (const char *dbfile, const char *sigfile)
{
  signed long diff;
  struct header_s
  {
    unsigned int magic;
    unsigned long timestamp;
#ifdef HAVE_UNSIGNED_LONG_LONG
    unsigned long long nentry;
#else
    unsigned long nentry;
#endif
  } header;
  struct read_entry_s
  {
    int checked;
    struct stat filestat;
    struct mesh_s meshes;
    unsigned int namelen;
    unsigned int hashval;
  } readbuf;
  struct hash_entry_s *entry, *ptr, *stop;
  gzFile fin;

  if (checkSignature (dbfile, sigfile, EICS_PUB_KEY))
    {
      fprintf (stderr,
	       "Database signature does not match database contents, please run with -i to initialize\n");
      return -1;
    }

  if (!(fin = gzopen (dbfile, "rb")))
    return -1;

  if (gzread (fin, &header, sizeof (header)) != sizeof (header))
    {
      gzclose (fin);
      return -1;
    }
  if (header.magic != DBMAGICNUM)
    {
      fprintf (stderr,
	       "Invalid magic in database start - please run with -i to initialize\n");
      gzclose (fin);
      return -1;
    }

  if ((diff = header.timestamp - time ((time_t *) NULL)) > 0)
    {
      fprintf (stderr,
	       "Header timestmap in future. Your clock is skewed. Difference is %lu\n",
	       diff);
    }

  while (!gzeof (fin))
    {
      if (gzread (fin, &readbuf, sizeof (readbuf)) != sizeof (readbuf))
	{
	  break;
	}
      entry = malloc (sizeof (struct hash_entry_s));
      memset (entry, 0, sizeof (struct hash_entry_s));
      memcpy (entry, &readbuf, sizeof (readbuf));
      /* Now we need to extract the tailing filename */
      entry->name = malloc (sizeof (char) * (entry->namelen + 1));
      if (gzread (fin, entry->name, sizeof (char) * (entry->namelen + 1)) !=
	  sizeof (char) * (entry->namelen + 1))
	{
	  fprintf (stderr, "Database corrupt - run with -i to initialize\n");
	  gzclose (fin);
	  return -1;
	}
      entry->checked = 0;
      /* Then, we add it to the hash table. */
      if (!hash_table.entries[entry->hashval])
	{
	  hash_table.entries[entry->hashval] = entry;
	}
      else
	{
	  ptr = hash_table.entries[entry->hashval];
	  while (ptr)
	    {
	      stop = ptr;
	      ptr = ptr->next;
	    }
	  stop->next = entry;
	}
      hash_table.nentry++;
    }
  gzclose (fin);
  if (hash_table.nentry != header.nentry)
    {
      printf ("Database inconsistent - missing entries\n");
      return -1;
    }
  return 0;
}

/* This is the longest process, we look out every unchecked item
 * and remove them.
 * Reporting is also done
 */
int
pruneDatabase (void)
{
  int i, rv;
  struct hash_entry_s *ptr, *prev, *next;
  rv = 0;
  for (i = 0; i < HASHTABLE_SIZE; i++)
    {
      if ((ptr = hash_table.entries[i]))
	{
	  prev = NULL;
	  while (ptr)
	    {
	      next = ptr->next;
	      /* Hasn't been checked during the other processes - remove */
	      if (!ptr->checked)
		{
		  fprintf (stderr, "Missing file %s\n", ptr->name);
		  rv = 1;
		  /*  Update pointers */
		  if (prev)
		    {
		      prev->next = ptr->next;
		    }
		  else
		    {
		      hash_table.entries[i] = ptr->next;
		    }
		  memset (ptr->name, 0, sizeof (char) * (ptr->namelen + 1));
		  free (ptr->name);
		  /* Delete info from memory */
		  memset (ptr, 0, sizeof (struct hash_entry_s));
		  /* Remove pointer */
		  free (ptr);
		  ptr = NULL;
		  hash_table.nentry--;
		}
	      prev = ptr;
	      ptr = next;
	    }
	}
    }
  /* IF and only IF something was changed, this will be -1 */
  return rv;
}
