/*
 * Policy reader
 *
 * (c) Aki Tossavainen 2004

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include "eics.h"

struct policy_s
{
  const char *regmask;
  const pcre *regex;
  const pcre_extra *study;
  const char *fmask;
  int flags;
} *policy;

size_t npolicy;

void
initPolicy (void)
{
  policy = NULL;
  npolicy = 0;
}

int
flagsToInt (const char *cflags)
{
  int flags = 0;
  for (; *cflags; cflags++)
    {
      switch (*cflags)
	{
	case 'a':
	  flags |= POLICY_ACCESS_TIME;
	  break;
	case 'c':
	  flags |= POLICY_CREATION_TIME;
	  break;
	case 'e':
	  flags |= POLICY_MODIFY_TIME;
	  break;
	case 'g':
	  flags |= POLICY_GROUP;
	  break;
	case 'h':
	  flags |= POLICY_MESH;
	  break;
	case 'i':
	  flags |= POLICY_INODE;
	  break;
	case 'l':
	  flags |= POLICY_MAYGROW;
	  break;
	case 'm':
	  flags |= POLICY_PERMISSIONS;
	  break;
	case 'n':
	  flags |= POLICY_NLINKS;
	  break;
	case 'u':
	  flags |= POLICY_OWNER;
	  break;
	case 's':
	  flags |= POLICY_SIZE;
	  break;
	case 'X':
	  flags |= POLICY_EXCLUDE;
	  break;
	case 'D':
	  flags |= POLICY_NOSUBDIRS;
	  break;
	default:
	  return -1;
	}
    }
  return (flags ? flags : -1);
}

void
endAtCrOrLf (char *line)
{
  char *ptr;
  if ((ptr = mb_strchr (line, '\n')))
    *ptr = '\0';
  if ((ptr = mb_strchr (line, '\r')))
    *ptr = '\0';
}

pcre **
createMasks ()
{
  const char *errptr;
  int errloc;
  pcre **rv = malloc (sizeof (pcre *) * 2);
  rv[0] = pcre_compile ("^\\s*?#", 0, &errptr, &errloc, NULL);
  rv[1] =
    pcre_compile ("^([^\\s]*)\\s([^\\s]*).*", 0, &errptr, &errloc, NULL);
  return rv;
}

int
getFirstPolicyFlags (void)
{
  return policy[0].flags;
}

const char *
getFirstPolicyDir (void)
{
  return policy[0].fmask;
}

int
isPolicyDefined (const char *mask, int regex)
{
  int i;
  for (i = 0; i < npolicy - 1; i++)
    {
      if ((regex) && (policy[i].regmask))
	{
	  if (!strcmp (mask, policy[i].regmask))
	    return -1;
	}
      else if (policy[i].fmask)
	{
	  if (!strcmp (mask, policy[i].fmask))
	    return -1;
	}
    }
  return 0;
}

int
addNewPolicy (int line, const char *mask, const char *flags)
{
  const char *pcre_error;
  int pcre_errloc;

  policy = realloc (policy, sizeof (struct policy_s) * (++npolicy));
  memset (policy + npolicy - 1, 0, sizeof (struct policy_s));
  if ((policy[npolicy - 1].flags = flagsToInt (flags)) == -1)
    {
      fprintf (stderr, "Line %d: Uknown flag(s)\n", line);
      return -1;
    }
  /* pcre masks start with worm */
  if ((*mask) == '~')
    {
      if (npolicy == 1)
	{
	  fprintf (stderr,
		   "Line %d: First policy cannot be a regular expression\n",
		   line);
	  return -1;
	}
      if (isPolicyDefined (mask + 1, 1))
	{
	  fprintf (stderr, "Line %d: Duplicate policy '%s'\n", line,
		   mask + 1);
	  return -1;
	}
      if (!
	  (policy[npolicy - 1].regex =
	   pcre_compile (mask + 1, 0, &pcre_error, &pcre_errloc, NULL)))
	{
	  fprintf (stderr, "Line %d: %s at %d\n", line, pcre_error,
		   pcre_errloc);
	  return -1;
	}
      policy[npolicy - 1].study =
	pcre_study (policy[npolicy - 1].regex, 0, &pcre_error);
      if (pcre_error)
	{
	  fprintf (stderr, "Line %d: %s\n", line, pcre_error);
	  return -1;
	}
      policy[npolicy - 1].regmask = strdup (mask + 1);
    }
  else
    {
      if ((*mask) != '/')
	{
	  fprintf (stderr, "Line %d: Non-PCRE masks must begin with /", line);
	  return -1;
	}
      if (isPolicyDefined (mask, 0))
	{
	  fprintf (stderr, "Line %d: Duplicate policy '%s'\n", line, mask);
	  return -1;
	}
      policy[npolicy - 1].fmask = strdup (mask);
    }
  return 0;
}

void
freemasks (pcre ** masks)
{
  pcre_free (masks[0]);
  pcre_free (masks[1]);
  free (masks);
}

int
patternmatch (const char *pattern, const char *subject)
{
  for (; (*pattern) && (*subject); subject++)
    {
      switch (*pattern)
	{
	case '*':
	  if (*(pattern + 1) == '\0')
	    return 0;
	  if (*(pattern + 1) == *(subject))
	    {
	      pattern++;
	      break;
	    }
	  continue;
	case '?':
	  if (*(subject + 1) == '\0')
	    return 0;
	  break;
	}
      if (*pattern != *subject)
	return (*pattern) - (*subject);
      pattern++;
    }
  if ((*pattern) != (*subject))
    return (*pattern) - (*subject);
  return 0;
}

int
matchPolicy (const char *file, size_t flen, int oldflags)
{
  int i;
  for (i = 0; i < npolicy; i++)
    {
      if (policy[i].regex)
	{
	  if (pcre_exec
	      (policy[i].regex, policy[i].study, file, flen, 0, PCRE_ANCHORED,
	       NULL, 0) == 0)
	    return policy[i].flags;
	}
      else
	{
	  if (!patternmatch (policy[i].fmask, file))
	    return policy[i].flags;
	}
    }
  return oldflags;
}

int
loadPolicy (const char *policyfile, const char *policysigfile, int doUpdate)
{
  int *ovec, osize, line, matches;
  pcre **masks;
  char buffer[4097];
  const char *filepattern, *flags;
  FILE *in;
  if (checkSignature (policyfile, policysigfile, EICS_PUB_KEY))
    {
      fprintf (stderr, "Policy file contents do not match signature!\n");
      if (doUpdate)
	{
	  fprintf (stderr,
		   "If you know that these changes were made by you\n");
	  fprintf (stderr, "verify changes with your password. Otherwise\n");
	  fprintf (stderr,
		   "press CTRL+C now and go investigate the problem.\n");
	  if (writeSignature (policyfile, policysigfile, EICS_PRIV_KEY))
	    {
	      return -1;
	    }
	}
      else
	{
	  fprintf (stderr, "Will not process filesystem\n");
	  return -1;
	}
    }
  if (!(in = fopen (policyfile, "r")))
    {
      return -1;
    }

  masks = createMasks ();
  ovec = malloc (sizeof (int) * 15);
  osize = 15;
  /* check signature!!!!!!! */
  line = 0;
  while (!feof (in))
    {
      memset (buffer, 0, sizeof (buffer));
      line++;
      fgets (buffer, 4096, in);
      endAtCrOrLf (buffer);

      if (*buffer)
	{
	  if (pcre_exec
	      (masks[0], NULL, buffer, strlen (buffer), 0, PCRE_ANCHORED,
	       NULL, 0) > -1)
	    {
	      continue;
	    }
	  else
	    {
	      if ((matches =
		   pcre_exec (masks[1], NULL, buffer, strlen (buffer), 0, 0,
			      ovec, osize)) > 0)
		{
		  pcre_get_substring (buffer, ovec, matches, 1, &filepattern);
		  pcre_get_substring (buffer, ovec, matches, 2, &flags);
		  if (addNewPolicy (line, filepattern, flags))
		    {
		      freemasks (masks);
		      fclose (in);
		      return -1;
		    }
		}
	    }
	}
    }
  freemasks (masks);
  fclose (in);
  if (npolicy == 0)
    {
      fprintf (stderr,
	       "No policy lines were found - please add at least one policy\n");
      return -1;
    }
  if (addNewPolicy (0, EICS_DB_FILE, "mnug"))
    return -1;
  if (addNewPolicy (0, EICS_DB_FILE_SIG, "mnug"))
    return -1;
  if (addNewPolicy (0, EICS_DB_FILE_TMP, "X"))
    return -1;
  return 0;
}
