/*
 * Processes filesystem entries per defined policy.
 *
 * (c) Aki Tossavainen 2004

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */
#include "eics.h"

/* For non-regular files - skips meshing .. */
int
processNonRegularEntry (char *entryname, struct stat *statent, int policy)
{
  int violations;
  struct hash_entry_s *entry, newEntry;
  memset (&newEntry, 0, sizeof (newEntry));
  violations = 0;
  newEntry.checked = 1;
  newEntry.name = entryname;
  newEntry.namelen = strlen (entryname);
  memcpy (&newEntry.filestat, statent, sizeof (struct stat));
  if ((entry = findEntryFromHashTable (entryname)))
    {
      violations = checkViolations (entry, &newEntry, policy);
      if (violations)
	{
	  fprintf (stderr, "%s:\n", entryname);
	  printViolations (entry, &newEntry, violations);
	}
    }
  else
    {
      fprintf (stderr, "New entry %s\n", entryname);
      violations = 1;
    }
  updateEntryToHashTable (&newEntry);
  return violations;
}

/* Normal processing */
int
processRegularEntry (char *entryname, struct stat *statent, int policy)
{
  int violations;
  struct hash_entry_s *entry, newEntry;
  memset (&newEntry, 0, sizeof (newEntry));
  violations = 0;
  newEntry.checked = 1;
  newEntry.name = entryname;
  newEntry.namelen = strlen (entryname);
  memcpy (&newEntry.filestat, statent, sizeof (struct stat));
  if (policy & POLICY_MESH)
    {
      meshFile (entryname, &newEntry.meshes);
    }
  if ((entry = findEntryFromHashTable (entryname)))
    {
      violations = checkViolations (entry, &newEntry, policy);
      if (violations)
	{
	  fprintf (stderr, "%s:\n", entryname);
	  printViolations (entry, &newEntry, violations);
	}
    }
  else
    {
      fprintf (stderr, "New entry %s\n", entryname);
      violations = 1;
    }
  updateEntryToHashTable (&newEntry);
  return violations;
}

int
processDirectory (const char *directory, int defaultPolicy)
{
  struct stat statent;
  DIR *dir;
  struct dirent *entry;
  int currentPolicy, rv, newrv;
  unsigned int maxlen;
  char *entryname = NULL;

  rv = newrv = 0;

  if (!entryChecked (directory))
    {
      /* Make sure that this directory exists in the database */
      currentPolicy =
	matchPolicy (directory, strlen (directory), defaultPolicy);
      if (lstat (directory, &statent))
	{
	  printf ("Unable to stat directory %s\n", directory);
	  return 0;
	}
      newrv =
	processNonRegularEntry ((char *) directory, &statent, currentPolicy);
    }

  if (newrv)
    {
      rv = newrv;
    }
  /* Try to open directory */
  if (!(dir = opendir (directory)))
    {
      fprintf (stderr, "Unable to descend into %s\n", directory);
      return 0;
    }

  /* Start reading entries */
  while ((entry = readdir (dir)))
    {
      /* Skip these two for obvious reasons */
      if ((strcmp (entry->d_name, ".")) && (strcmp (entry->d_name, "..")))
	{
	  /* Reallocate our entryname pointer, why waste a good pointer? */
	  maxlen = strlen (directory) + NAMLEN (entry) + 1;
	  entryname = realloc (entryname, sizeof (char) * (maxlen + 1));
	  strncpy (entryname, directory, maxlen);
	  if (*(mb_strrchr (entryname, '/') + 1))
	    strncat (entryname, "/", maxlen - strlen (entryname));
	  strncat (entryname, entry->d_name, maxlen - strlen (entryname));
	  *(entryname + maxlen) = '\0';
	  currentPolicy =
	    matchPolicy (entryname, strlen (entryname), defaultPolicy);

	  if ((currentPolicy & POLICY_EXCLUDE) != POLICY_EXCLUDE)
	    {

	      if (lstat (entryname, &statent))
		{
		  printf ("Unable to stat %s\n", entryname);
		  continue;
		}

	      if (S_ISREG (statent.st_mode))
		{
		  newrv =
		    processRegularEntry (entryname, &statent, currentPolicy);
		}
	      else
		{
		  newrv =
		    processNonRegularEntry (entryname, &statent,
					    currentPolicy);
		}
	      if (newrv)
		{
		  rv = newrv;
		}
	      if ((S_ISDIR (statent.st_mode))
		  && (!(currentPolicy & POLICY_NOSUBDIRS)))
		{
		  newrv = processDirectory (entryname, currentPolicy);
		}
	    }
	  /* To preserve return value */
	  if (newrv)
	    {
	      rv = newrv;
	    }
	}
      if (newrv)
	{
	  rv = newrv;
	}
    }
  closedir (dir);
  return rv;
}
