/*
 * Digital signature handling functions
 *
 * (c) Aki Tossavainen 2004

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include "eics.h"

/* Returns the public DSA key */
EVP_PKEY *
openDSAPublicKey (const char *pubkeyfile)
{
  BIO *keyIn;
  EVP_PKEY *rv;

  if (!(keyIn = BIO_new_file (pubkeyfile, "r")))
    return NULL;

  rv = PEM_read_bio_PUBKEY (keyIn, NULL, NULL, NULL);

  BIO_free (keyIn);

  return rv;
}

/* Returns the private DSA key */
EVP_PKEY *
openDSAPrivateKey (const char *privatekeyfile)
{
  int i;
  BIO *keyIn;
  EVP_PKEY *rv = NULL;

  /* Fixed feature: now asks three time before giving up */
  for (i = 0; i < 3 && rv == NULL; i++)
    {
      if (!(keyIn = BIO_new_file (privatekeyfile, "r")))
	return NULL;
      rv = PEM_read_bio_PrivateKey (keyIn, NULL, pass_cb, NULL);
      BIO_free (keyIn);
    }
  return rv;
}

/* Writes the keypair into a file */
int
writeDSAKeyPair (DSA * dsa, const char *publickeyfile,
		 const char *privatekeyfile)
{
  BIO *out;

  if (!(out = BIO_new_file (publickeyfile, "w")))
    {
      printf ("Failed to write public key into %s\n", publickeyfile);
      return -1;
    }
  if (!PEM_write_bio_DSA_PUBKEY (out, dsa))
    {
      printf ("Failed to write public key into %s\n", publickeyfile);
      BIO_free (out);
      return -1;
    }
  BIO_free (out);

  if (!(out = BIO_new_file (privatekeyfile, "w")))
    {
      printf ("Failed to write private key into %s\n", privatekeyfile);
      return -1;
    }
  if (!PEM_write_bio_DSAPrivateKey
      (out, dsa, EVP_bf_cbc (), NULL, 0, pass_cb, "x"))
    {
      printf ("Failed to write private key into %s\n", privatekeyfile);
      BIO_free (out);
      return -1;
    }
  BIO_free (out);
  return 0;
}

/* Generates a brand new keypair */
int
generateNewKeys (const char *publickeyfile, const char *privatekeyfile)
{
  DSA *new;

  printf ("Generating DSA keypair...");
  fflush (NULL);

  if (!
      (new =
       DSA_generate_parameters (DSA_KEY_STRENGTH, NULL, 0, NULL, NULL, NULL,
				NULL)))
    {
      printf ("failed\n");
      ERR_print_errors_fp (stdout);
      return -1;
    }
  if (!DSA_generate_key (new))
    {
      printf ("failed\n");
      ERR_print_errors_fp (stdout);
      DSA_free (new);
      return -1;
    }
  printf ("success!\n");
  if (!writeDSAKeyPair (new, publickeyfile, privatekeyfile))
    {
      printf ("New keypair written to disk\n");
    }
  DSA_free (new);
  return 0;
}

/* Checks for signature validity */
int
checkSignature (const char *file, const char *sigfile, const char *pubkeyfile)
{
  BIO *bin = NULL;
  BIO *sigin = NULL;
  int rv;
  unsigned char tmpbuf[1024];
  unsigned char *sigbuf = NULL;
  unsigned int siglen = 0;
  EVP_PKEY *pubkey = NULL;
  EVP_MD_CTX *ctx;

  if (!(sigin = BIO_new_file (sigfile, "r")))
    {
      return -1;
    }

  if (!(bin = BIO_new_file (file, "r")))
    {
      BIO_free (sigin);
      return -1;
    }
  ctx = EVP_MD_CTX_create ();

  EVP_VerifyInit (ctx, EVP_dss1 ());

  /* Here we digest and digest and digest... */
  while ((rv = BIO_read (bin, tmpbuf, sizeof (tmpbuf))) > 0)
    EVP_VerifyUpdate (ctx, tmpbuf, rv);
  memset (tmpbuf, 0, sizeof (tmpbuf));

  /* Verify signature */
  if ((pubkey = openDSAPublicKey (pubkeyfile)))
    {
      siglen = EVP_PKEY_size (pubkey);
      sigbuf = malloc (siglen);
      BIO_read (sigin, sigbuf, siglen);
      rv = (EVP_VerifyFinal (ctx, sigbuf, siglen, pubkey) == 1 ? 0 : -1);
    }
  else
    {
      rv = -1;
    }

  /* Clean everything up */
  EVP_MD_CTX_destroy (ctx);
  if (pubkey)
    memset (pubkey, 0, sizeof (EVP_PKEY));
  if (sigbuf)
    {
      memset (sigbuf, 0, siglen);
      free (sigbuf);
    }
  BIO_free (bin);
  BIO_free (sigin);
  return rv;
}

/* Writes new signature to disk */
int
writeSignature (const char *file, const char *sigfile,
		const char *privkeyfile)
{
  EVP_MD_CTX *ctx = NULL;
  BIO *bin = NULL;
  BIO *bout = NULL;
  EVP_PKEY *privkey = NULL;
  int rv;
  unsigned char tmpbuf[1024];
  unsigned char *sigbuf = NULL;
  unsigned int siglen = 0;

  if (!(bin = BIO_new_file (file, "rb")))
    {
      return -1;
    }

  ctx = EVP_MD_CTX_create ();

  EVP_SignInit (ctx, EVP_dss1 ());

  /* Digest */
  while ((rv = BIO_read (bin, tmpbuf, sizeof (tmpbuf))) > 0)
    EVP_SignUpdate (ctx, tmpbuf, rv);
  memset (tmpbuf, 0, sizeof (tmpbuf));

  /* Create and write signature */
  if ((privkey = openDSAPrivateKey (privkeyfile)))
    {
      siglen = EVP_PKEY_size (privkey);
      sigbuf = malloc (siglen);

      rv = (EVP_SignFinal (ctx, sigbuf, &siglen, privkey) ? 0 : -1);
      if (!(bout = BIO_new_file (sigfile, "w")))
	{
	  rv = -1;
	}
      else
	{
	  BIO_write (bout, sigbuf, siglen);
	}
    }
  else
    {
      rv = -1;
    }

  /* OK, we're done 
   * Clean up the trash 
   */
  EVP_MD_CTX_destroy (ctx);
  if (privkey)
    memset (privkey, 0, sizeof (EVP_PKEY));
  if (sigbuf)
    {
      memset (sigbuf, 0, EVP_MAX_MD_SIZE);
      free (sigbuf);
    }
  BIO_free (bin);
  if (bout)
    BIO_free (bout);
  return rv;
}
